% pool using work dealing
-module(dpool).
-export([init_pool/6,pmap/3,preduce/4]).


% create worker for `Pool' computing `Function'
init_worker(Pool, Function) ->
  spawn(fun ()-> worker(Pool, Function) end).

worker(Pool, Function) ->
  receive {Pool, Data} ->     % assignment from pool
    Result = Function(Data),  % compute work
    Pool ! {self(), Result},  % send result to pool
    worker(Pool, Function)    % back to idle
  end.


% pool state:
%  work: pending work items
%  busy: list of busy workers
%  idle: list of idle workers
-record(pool, {work, busy, idle}).

% work completed, no busy workers: return result
pool(#pool{work = [], busy = []}, 
     _Split, _Join, Result) ->
  Result;

% work pending, some idle workers: assign work
pool(Pool = #pool{work = Work = [_|_], % matches if Work not empty
                  busy = Busy, 
                  idle = [Worker|Idle]}, 
     Split, Join, Result) ->
    {Chunk, Remaining} = Split(Work),  % split pending work
    Worker ! {self(), Chunk},          % send chunk to worker
    pool(Pool#pool{work = Remaining,
                   busy = [Worker|Busy],
                   idle = Idle},
         Split, Join, Result);

% work completed or no idle workers: wait for results
pool(Pool = #pool{busy = Busy, idle = Idle}, 
     Split, Join, Result) ->
         % get result from worker
    receive {Worker, PartialResult} -> ok end,  
         % join worker's result and current result
    NewResult = Join(PartialResult, Result),
    pool(Pool#pool{busy = lists:delete(Worker, Busy),
                   idle = Idle ++ [Worker]}, 
         Split, Join, NewResult).

% start a pool to compute `Function' on `Work' using `N' process workers
% Split: Work -> {Unit, RemainingWork}  (splits one unit of work from input)
% Join: Result x Result -> Result       (combines two partial results)
% Initial: initial result
init_pool(Function, Work, Split, Join, Initial, N) ->
  Pool = self(),
     % spawn N workers for the same pool
  Workers = [init_worker(Pool, Function) || _ <- lists:seq(1, N)],
  [link(W) || W <- Workers], % link workers to pool
     % initially all work is pending, all workers are idle
  pool(#pool{work = Work, busy = [], idle = Workers}, 
       Split, Join, Initial).


% parallel map using a pool with N processes
%  F: mapped function
%  L: list
% Order of results is nondeterministic
pmap(F, L, N) -> init_pool(F,  % function to be mapped
   L,                          % work: list to be mapped
   fun ([H|T]) -> {H, T} end,  % split: take first element
   fun (R,Res) -> [R|Res] end, % join: cons with list
   [], N).


% parallel map using a pool with N processes
%  F: mapped function
%  L: list
% Does not guarantee order of results
% Works only if F(I, X) = F(X, I) for all X
preduce(F, I, L, N) ->
  init_pool(fun ({X,Y}) -> F(X,Y) end, % so that a chunk is a pair
            L,   % split: take first two elements
            fun (W) -> chunk_two(I, W) end,  
            F,   % join: folding function!
            I, N).

% take two elements, if possible
chunk_two(_, [F|[S|R]]) -> {{F,S}, R};
chunk_two(I, [F|R]) -> {{F,I}, R}.
 
